#ifndef GC_POINTER_H
#define GC_POINTER_H

#include "gc/link.h"

namespace gc {

/*
refcounting is cool as long as no one ends up mutually pointing to eachother when
they get spun off from the rest of the world... then they become a memory leak...

so how about that mark-and-sweep?
what does it require?

1) keep track of all your smart-ptr'd objects (that means a singleton, and that means C++,
	and that means this is going to end up in base)
2) keep one as the main one -- or a separate list as the main ones --
3) keep track of all smart ptrs from between those objects (for graph traversal) within each obj
*/

class Object;
class GarbageCollectSystem;

class PointerBase : public Link {
	//make the ctors protected so only Object's can instanciate them (friend with)
	friend class Object;
protected:
	Object *owner;
	Object *p;
	PointerBase(Object *_owner, Object *_p);
	PointerBase(const PointerBase &o);
	virtual ~PointerBase();
public:
	Object *ptrBase() { return p; }
	PointerBase &operator=(Object *_p);
	PointerBase &operator=(const PointerBase &_p);
};

//assert T is an instance of PointerBase ... so that the Object's ptrs list can hold them
template<typename T>
class Pointer : public PointerBase {
	friend class Object;
	typedef PointerBase super;
	typedef T *Tptr;
protected:
/*
do we need to hold the type-specific pointer separately? (more memory / more data to synchronize...)
or should we just dynamic_cast every time? (more cycles)
how about safety too?  should we pass it a dereference from the owner's class to the ptr?
	and store a chain of those?  for for type safety's sake?
here's another deal: confusion in initialization...
how about we don't allow initialization...
but instead have all pointer creation come from a member method of gc::Object?
that'd ensure correct attachment..
*/
	Pointer(Object *_owner, Tptr _p) : super(_owner, _p) {}
public:

	Tptr ptr() { return dynamic_cast<Tptr>(p); }
	const Tptr ptr() const { return const_cast<const Tptr>(dynamic_cast<Tptr>(p)); }

	Tptr operator->() { return dynamic_cast<Tptr>(p); }
	const Tptr operator->() const { return const_cast<const Tptr>(dynamic_cast<Tptr>(p)); }
	
	template<typename U>
	bool operator==(const U *_p) const {
		return p == dynamic_cast<const Object *>(_p);
	}

	template<typename U>
	bool operator!=(const U *_p) const { return !operator==<U>(_p); }

	template<typename U>
	bool operator==(const Pointer<U> &_p) const { return operator==<U>(_p.ptr()); }

	template<typename U>
	bool operator!=(const Pointer<U> &_p) const { return operator!=<U>(_p.ptr()); }

	//TODO - a dynamic_cast will return NULL if the cast is bad
	//but this returned won't tell the user whether the cast failed or the pointer was NULL to begin with
	//soo.... how about a bad-cast exception thrown?
	template<typename U>
	operator U*() { return dynamic_cast<U*>(p); }

	template<typename U>
	operator const U*() const { return dynamic_cast<U*>(p); }
	
	operator bool() const { return !!p; }
	
	Pointer &operator=(Tptr _p) { super::operator=(dynamic_cast<Object*>(_p)); return *this; }
	Pointer &operator=(const Pointer<T> &_p) { super::operator=(_p); return *this; }
};

//member pointers MUST be ctor'd with their owners
//if it's not a memeber... meh...
template<typename T>
class GlobalPointer : public Pointer<T> {
	typedef Pointer<T> super;
	friend class GarbageCollectSystem;
	typedef T *Tptr;
protected:
	GlobalPointer(T *_p) : super(0, _p) {}
public:
	GlobalPointer &operator=(Tptr _p) { super::operator=(_p); return *this; }
	GlobalPointer &operator=(const GlobalPointer &_p) { super::operator=(_p); return *this; }

};

};

#endif	//GC_POINTER_H
